/*
 * drivers/char/sid.c - SUZAKU I/O Boards -A/D- Driver
 *
 * Copyright (C) 2006-2007 Atmark Techno, Inc.
 * Chris McHarg <chris (at) atmark-techno.com>
 * Masahiro Hanada <hanada (at) atmark-techno.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2,
 * as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * FIXME:
 *  - the driver assume that dma_addr_t is u32.
 */

//#define _DEBUG

#include <linux/version.h>
#include <linux/config.h>
#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/slab.h>
#include <linux/fs.h>
#include <linux/errno.h>
#include <linux/types.h>
#include <linux/ioport.h>
#include <linux/interrupt.h>
#include <linux/wait.h>
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,5,0)
#include <linux/platform_device.h>
#include <linux/device.h>
#include <linux/dmapool.h>
#include <linux/cdev.h>
#include <linux/sysfs.h>
typedef struct dma_pool struct_dma_pool;
typedef struct device struct_device;
#else
#define dev_err(dev, format, arg...) \
        printk(KERN_ERR "%s: " format, DRIVER_NAME, ##arg)
#define dev_warn(dev, format, arg...) \
        printk(KERN_WARNING "%s: " format, DRIVER_NAME, ##arg)
#ifdef _DEBUG
#define dev_err(dev, format, arg...) \
        printk(KERN_ERR "%s(%d): " format, __FILE__, __LINE__, ##arg)
#define dev_warn(dev, format, arg...) \
        printk(KERN_WARNING "%s(%d): " format, __FILE__, __LINE__, ##arg)
#define dev_info(dev, format, arg...) \
        printk(KERN_INFO "%s(%d): " format, __FILE__, __LINE__, ##arg)
#define dev_dbg(dev, format, arg...) \
        printk(KERN_DEBUG "%s(%d): " format, __FILE__, __LINE__, ##arg)
#else
#define dev_info(dev, format, arg...)
#define dev_dbg(dev, format, arg...)
#endif
typedef int gfp_t;
#include <linux/pci.h>
#include <asm/pci.h>
typedef struct pci_pool struct_dma_pool;
typedef struct pci_dev struct_device;
#define dma_pool_create(name, dev, size, align, allocation) \
        pci_pool_create((name), (dev), (size), (align), (allocation), \
			GFP_KERNEL|GFP_DMA)
#define dma_pool_destroy(pool) \
        pci_pool_destroy(pool)
#define dma_pool_alloc(pool, flags, handle) \
        pci_pool_alloc((pool), (flags), (handle))
#define dma_pool_free(pool, vaddr, addr) \
        pci_pool_free((pool), (vaddr), (addr))
#define MODULE_VERSION(ver_str)
#define ALIGN(x,a) (((x)+(a)-1)&~((a)-1))
#define wait_for_completion_interruptible(x) wait_for_completion(x)
#define iminor(inode) MINOR((inode)->i_rdev)
static inline void *kcalloc(size_t nmemb, size_t size, int flags)
{
        size_t total = nmemb * size;
        void *mem = kmalloc(total, flags);
        if (mem)
                memset(mem, 0, total);
        return mem;
}
#define kzalloc(size, flags) kcalloc(1, (size), (flags))
#endif
#include <asm/uaccess.h>
#include <asm/io.h>
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
#include <asm/suzaku.h>
#endif
#include <asm/suzaku_sid.h>
#include <asm/delay.h>

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
#define NO_ALLOC_IN_INTERRUPT
#endif

#ifdef _DEBUG
#undef KERN_DEBUG
#define KERN_DEBUG KERN_ERR
#endif

#define DRIVER_NAME			"sid"
#define DRIVER_VERSION                  "v1.0.1"
#define DRIVER_DESC                     "SUZAKU I/O A/D Board Driver"

/**
 * Register addresses and setting values
 */
#define REG_CONTROL			(0x0000)
#define REG_CONTROL_SAMPLE		(1 << 0)
#define REG_CONTROL_DMA			(1 << 1)
#define REG_CONTROL_HW_RESET		(1 << 31)

#define REG_SAMPLE_CLOCK		(0x0060)
#define REG_SAMPLE_RATE			(0x0064)
#define REG_DMA_ADDRESS			(0x0080)
#define REG_DMA_LENGTH			(0x0084)

#define REG_STATUS			(0x0010)
#define REG_STATUS_OVERRUN		(1 << 0)
#define REG_STATUS_DMA_DONE		(1 << 1)
#define REG_STATUS_HW_RESET		(1 << 31)

#define REG_STATUS_CLEAR		(0x0014)
#define REG_STATUS_CLEAR_OVERRUN	(1 << 0)
#define REG_STATUS_CLEAR_DMA_DONE	(1 << 1)
#define REG_STATUS_CLEAR_HW_RESET	(1 << 31)

#define REG_INTR_E			(0x0040)
#define REG_INTR_E_OVERRUN		(1 << 0)
#define REG_INTR_E_DMA_DONE		(1 << 1)
#define REG_INTR_E_HW_RESET		(1 << 31)

#define REG_ID				(0x00FC)

#define REG_ADC_CHANNELS                (0x00A0)
#define REG_ADC_RESOLUTION              (0x00A4)

#define SAMPLE_RATE_MAX                 (150000)
#define BYTES_PER_CHANNEL               (2)
#define BYTES_PER_UNIT(ch)		((ch)*BYTES_PER_CHANNEL)

static unsigned int default_sample_rate = 10000;
module_param(default_sample_rate, uint, 10000);

MODULE_PARM_DESC(default_sample_rate, "Default sample rate");

struct dma_block {
	struct list_head list;
	dma_addr_t handle;
	unsigned long len;
	unsigned long pos;
	/* data must be align to 4 byte */
	u8 data[0];
};

static unsigned long get_block_len(struct dma_block *block)
{
	return block->len - block->pos;
}

static int is_block_empty(struct dma_block *block)
{
	return !(block->len - block->pos);
}

struct sid {
	struct_device *dev;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,5,0)
	struct cdev cdev;
#endif
	struct_dma_pool *dma_pool;
	dev_t devno;

	spinlock_t dma_blocks_lock;

	struct list_head dma_blocks;
	struct list_head read_blocks;
	struct dma_block *next_block;
	struct dma_block *current_block;
	struct dma_block *partial_block;

	unsigned long bytes_per_block;

	struct completion complete;

	int irq;
	void __iomem *addr;
	unsigned long region_start;
	unsigned long region_len;

	unsigned int channels;
	unsigned int resolution;
	unsigned int sample_clock;
	u32 reg_sample_rate;
	unsigned int sample_rate;
	struct {
		unsigned long rx, overrun;
	} stat;
};

#define MAX_NR_AD_BOARDS (1)
static struct sid *sids[MAX_NR_AD_BOARDS];

/**
 * _write_reg - writes to registers
 * @val: the value to write
 * @offset: the offset from mapped memory of the target register
 */
static void _write_reg(struct sid *sid, u32 offset, u32 val)
{
	__raw_writel(val, sid->addr + offset);
}

/**
 * _read_reg - reads from registers
 * @offset: the offset from mapped memory of the target register
 */
static u32 _read_reg(struct sid *sid, u32 offset)
{
	return __raw_readl(sid->addr + offset);
}

static void set_dma_address(struct sid *sid, dma_addr_t addr)
{
	_write_reg(sid, REG_DMA_ADDRESS, addr + sizeof(struct dma_block));
}

static void set_dma_length(struct sid *sid, u32 count)
{
	sid->bytes_per_block = count * BYTES_PER_CHANNEL * sid->channels;
	_write_reg(sid, REG_DMA_LENGTH, count);
}

static unsigned long get_bytes_per_block(struct sid *sid)
{
	return sid->bytes_per_block;
}

static void dma_start(struct sid *sid)
{
	u32 val = _read_reg(sid, REG_CONTROL);
	_write_reg(sid, REG_CONTROL, val | REG_CONTROL_DMA);
}

static int set_sample_rate(struct sid *sid, unsigned int rate)
{
	if (rate == sid->sample_rate)
		return 0;
	if (rate > SAMPLE_RATE_MAX)
		return -EINVAL;

	if (sid->sample_clock % rate)
		dev_warn(sid->dev, "given sample rate %uHz is not acceptable. " \
			 "using approximate value.\n", rate);

	dev_dbg(sid->dev, "%s\n", __FUNCTION__);
	dev_dbg(sid->dev, "  %u => %u\n", sid->sample_rate, rate);

	sid->reg_sample_rate = sid->sample_clock / rate;
	sid->sample_rate = rate;

	return 0;
}

static struct dma_block *alloc_dma_block(struct sid *sid, gfp_t flags)
{
	dma_addr_t handle;
	struct dma_block *ret;

	ret = dma_pool_alloc(sid->dma_pool, flags, &handle);
	if (ret) {
		ret->handle = handle;
		ret->len = get_bytes_per_block(sid);
		ret->pos = 0;
/* 		if (!in_interrupt()) { */
/* 			dev_dbg(sid->dev, "%s\n", __FUNCTION__); */
/* 			dev_dbg(sid->dev, "  vaddr: %p\n", ret); */
/* 			dev_dbg(sid->dev, "  paddr: %#x\n", ret->handle); */
/* 			dev_dbg(sid->dev, "  len: %lu\n", ret->len); */
/* 			dev_dbg(sid->dev, "  pos: %lu\n", ret->pos); */
/* 		} */
	}
	return ret;
}

static void __free_dma_block(struct sid *sid, struct dma_block *block, char *from)
{
	if (block) {
/* 		if (!in_interrupt()) { */
/* 			dev_dbg(sid->dev, "%s(%s)\n", __FUNCTION__, from); */
/* 			dev_dbg(sid->dev, "  vaddr: %p\n", block); */
/* 			dev_dbg(sid->dev, "  paddr: %#x\n", block->handle); */
/* 			dev_dbg(sid->dev, "  len: %lu\n", block->len); */
/* 			dev_dbg(sid->dev, "  pos: %lu\n", block->pos); */
/* 		} */
		dma_pool_free(sid->dma_pool, block, block->handle);
	}
}

#define free_dma_block(sid, block) __free_dma_block(sid, block, __stringify(block))

#ifdef NO_ALLOC_IN_INTERRUPT
static inline void append_dma_block(struct sid *sid)
{
	if (!sid->current_block) {
		sid->current_block = alloc_dma_block(sid, GFP_ATOMIC);
#ifdef _DEBUG
		if (!sid->current_block)
			dev_dbg(sid->dev, "alloc current_block failed.\n");
#endif
	}
	if (sid->current_block) {
		if (!sid->next_block) {
			sid->next_block = alloc_dma_block(sid, GFP_ATOMIC);
#ifdef _DEBUG
			if (!sid->next_block)
				dev_dbg(sid->dev,
					"alloc next_block failed.\n");
#endif
		}
	}
}
#endif

static ssize_t sid_read(struct file *filp, char __user *buf, size_t count, loff_t *f_pos)
{
	struct sid *sid = filp->private_data;
	struct dma_block *block, *n;
	unsigned long flags;
	ssize_t ret = 0;

	dev_dbg(sid->dev, "%s\n", __FUNCTION__);
	dev_dbg(sid->dev, "  count = %d\n", count);
	dev_dbg(sid->dev, "  buf = %p\n", buf);

	while (list_empty(&sid->read_blocks)) {
		spin_lock_irqsave(&sid->dma_blocks_lock, flags);

		if (list_empty(&sid->dma_blocks)) {
#ifdef NO_ALLOC_IN_INTERRUPT
			append_dma_block(sid);
			if (!sid->current_block) {
				spin_unlock_irqrestore
				  (&sid->dma_blocks_lock, flags);
				return -ENOMEM;
			}
#endif
			spin_unlock_irqrestore(&sid->dma_blocks_lock, flags);
			dev_dbg(sid->dev, "  sleeping...\n");
			wait_for_completion_interruptible(&sid->complete);
			dev_dbg(sid->dev, "  wakeup\n");
			spin_lock_irqsave(&sid->dma_blocks_lock, flags);
		}
#ifdef NO_ALLOC_IN_INTERRUPT
		append_dma_block(sid);
#endif
		list_for_each_entry_safe(block, n, &sid->dma_blocks, list) {
			list_move_tail(&block->list, &sid->read_blocks);
		}

		spin_unlock_irqrestore(&sid->dma_blocks_lock, flags);
	}

	list_for_each_entry_safe(block, n, &sid->read_blocks, list) {
		unsigned long c;

		dev_dbg(sid->dev, "copy_to_user\n");
		dev_dbg(sid->dev, "  =%hx\n", *(u16*)block->data);

		c = min_t(unsigned long, count - ret, get_block_len(block));
		if (copy_to_user(buf + ret, &block->data[block->pos], c)) {
			/* we don't care about the current
			 * block if we failed. free it */
			list_del(&block->list);
			free_dma_block(sid, block);
			ret = -EFAULT;
			break;
		}

		dev_dbg(sid->dev, "  %#hx\n", *(u16*)buf);

		ret += c;
		sid->stat.rx += c;
		block->pos += c;
		dev_dbg(sid->dev, "  %lu bytes copied from block %p\n",
			c, block);

		if (is_block_empty(block)) {
			list_del(&block->list);
			free_dma_block(sid, block);
		}
#ifdef NO_ALLOC_IN_INTERRUPT
		spin_lock_irqsave(&sid->dma_blocks_lock, flags);
		append_dma_block(sid);
		spin_unlock_irqrestore(&sid->dma_blocks_lock, flags);
#endif
		if (count == ret)
			break;
	}

	dev_dbg(sid->dev, "  ret = %d\n", ret);

	return ret;
}

static void dbg_regdump(struct sid *sid)
{
	dev_dbg(sid->dev, " control: %#x\n", _read_reg(sid, REG_CONTROL));
	dev_dbg(sid->dev, "  status: %#x\n", _read_reg(sid, REG_STATUS));
	dev_dbg(sid->dev, "   clear: %#x\n", _read_reg(sid, REG_STATUS_CLEAR));
	dev_dbg(sid->dev, "  length: %#x\n", _read_reg(sid, REG_DMA_LENGTH));
}

static void hw_reset(struct sid *sid)
{
	dev_dbg(sid->dev, "reseting hardware...\n");
	_write_reg(sid, REG_CONTROL, REG_CONTROL_HW_RESET);
	while(! (_read_reg(sid, REG_STATUS) & REG_STATUS_HW_RESET))
		udelay(1000);
	_write_reg(sid, REG_STATUS_CLEAR, REG_STATUS_CLEAR_HW_RESET);

	dbg_regdump(sid);
}

static void cleanup_dma_mem(struct sid *sid)
{
	struct dma_block *dma_block, *read_block, *n;

	/* return all dma buffers to pool if any */
	if (sid->current_block)
		free_dma_block(sid, sid->current_block);

	if (sid->next_block)
		free_dma_block(sid, sid->next_block);

	list_for_each_entry_safe(dma_block, n, &sid->dma_blocks, list) {
		list_del(&dma_block->list);
		free_dma_block(sid, dma_block);
	}

	list_for_each_entry_safe(read_block, n, &sid->read_blocks, list) {
		list_del(&read_block->list);
		free_dma_block(sid, read_block);
	}

	/* remove dma pool, if any */
	if (sid->dma_pool)
		dma_pool_destroy(sid->dma_pool);

	sid->current_block = sid->next_block = NULL;
	sid->dma_pool = NULL;
}

#define UNIT_PER_PAGE(ch)   (PAGE_SIZE/BYTES_PER_UNIT(ch))
#define UNIT_PER_PAGE_ORDER (7) /* avoid unnecceraly calcuration for now */ 
#define SRATE_2_UNIT(x,ch)  (ALIGN((x),UNIT_PER_PAGE(ch))>>UNIT_PER_PAGE_ORDER) 
#define SRATE_2_BYTES(x,ch) (SRATE_2_UNIT((x),(ch))*BYTES_PER_UNIT(ch))

static int do_reset(struct sid *sid, unsigned int sample_rate)
{
	struct_dma_pool *pool;
	size_t size;
	int ret;

	/* first we reset the hardware */
	hw_reset(sid);

	/* then, cleanup all dma memory */
	cleanup_dma_mem(sid);

	/* first we reset the sample rate */
	ret = set_sample_rate(sid, sample_rate);
	if (ret) {
		pr_debug("Failed to set sample rate\n");
		return ret;
	}

	/* first we get dma length in unit from sample rate */
	set_dma_length(sid, SRATE_2_UNIT(sid->sample_rate, sid->channels));

	/* now we need to calculate the pool size from the sample
	 * rate and allocate dma pool */
	size = SRATE_2_BYTES(sid->sample_rate, sid->channels);
	if (size)
		size += sizeof(struct dma_block);
	pool = dma_pool_create("sid-0", sid->dev, size, 4, 0);
	if (!pool) {
		pr_debug("Failed to allocate DMA pool\n");
		return -ENOMEM;
	}
	sid->dma_pool = pool;

	dev_dbg(sid->dev, "dma pool created\n");
	dev_dbg(sid->dev, "  size: %#x\n", size);

	dev_dbg(sid->dev, "dma length\n");
	//dev_dbg(sid->dev, "  nr_unit: %lu units (%lu bytes)\n",
	//	nr_unit, get_bytes_per_block(sid));


	/* allocate two blocks, one for the current and another for
	 * the next to start working.  we can sleep here. */
	sid->current_block = alloc_dma_block(sid, GFP_ATOMIC|__GFP_WAIT);
	sid->next_block = alloc_dma_block(sid, GFP_ATOMIC|__GFP_WAIT);
	if (!sid->current_block || !sid->next_block)
		/* FIXME: let pool leak? */
		return -ENOMEM;

	dev_dbg(sid->dev, "check init val\n");
	dev_dbg(sid->dev, "  current(%p) %#hx\n", sid->current_block,
		*(u16 *)sid->current_block->data);
	dev_dbg(sid->dev, "  next(%p)    %#hx\n", sid->next_block,
		*(u16 *)sid->next_block->data);

	set_dma_address(sid, sid->current_block->handle);

	/* OK, everything is set, re-activate the hardware */
	_write_reg(sid, REG_SAMPLE_RATE, sid->reg_sample_rate);
	_write_reg(sid, REG_INTR_E, REG_INTR_E_DMA_DONE | REG_INTR_E_OVERRUN);
	_write_reg(sid, REG_CONTROL, REG_CONTROL_DMA | REG_CONTROL_SAMPLE);

	return 0;
}

/**
 * sid_ioctl - the ioctl implementation
 *
 * SID_IOC_GET_CHANNELS - gets the ad channels
 * SID_IOC_GET_RESOLUTION - gets the ad resolution
 * SID_IOC_SET_FREQ - sets the ad sample frequency
 * SID_IOC_RESET - resets the ad board (clears the fifos)
 */
static int sid_ioctl(struct inode *inode, struct file *filp,
		    unsigned int cmd, unsigned long arg)
{
	struct sid *sid = filp->private_data;
	unsigned int sample_rate = sid->sample_rate;
	int ret = 0;

	/* sanity checks */
	if (_IOC_TYPE(cmd) != SID_IOC_MAGIC)
		return -ENOTTY;
	if (_IOC_NR(cmd) > SID_IOC_MAXNR)
		return -ENOTTY;

	switch(cmd) {
	case SID_IOC_GET_CHANNELS:
		ret = put_user(sid->channels, (int *)arg);
		break;

	case SID_IOC_GET_RESOLUTION:
		ret = put_user(sid->resolution, (int *)arg);
		break;

	case SID_IOC_SET_FREQ:
		sample_rate = (unsigned int)arg;
		/* FALL THROUGH */

	case SID_IOC_RESET:
		ret = do_reset(sid, sample_rate);
		break;
	}

	return ret;
}

/**
 * sid_intr - handles ad related interrupts
 *
 * Does nothing if there's no room on the ad_buffer, which
 * will cause the fpga fifo to overrun.
 */
static irqreturn_t sid_intr(int irq, void *dev_id, struct pt_regs *regs)
{
	struct sid *sid = dev_id;
	u32 status = _read_reg(sid, REG_STATUS);

	if (!status)
		return IRQ_NONE;

/* 	dev_dbg(sid->dev, "got inter\n"); */
/* 	dbg_regdump(sid); */

	if (status & REG_STATUS_DMA_DONE) {
		unsigned long flags;

/* 		dev_dbg(sid->dev, "dma done\n"); */
/* 		dev_dbg(sid->dev, "  %lu bytes copied\n", sid->bytes_per_block); */
/* 		dev_dbg(sid->dev, "  -%hx\n", *(u16 *)sid->current_block->data); */

		spin_lock_irqsave(&sid->dma_blocks_lock, flags);
		list_add_tail(&sid->current_block->list, &sid->dma_blocks);
		spin_unlock_irqrestore(&sid->dma_blocks_lock, flags);

		sid->current_block = sid->next_block;
#ifndef NO_ALLOC_IN_INTERRUPT
		if (unlikely(!sid->current_block)) {
			/* this is the slow path */
			/* we don't have any other chance to allocate
			   a dma block we need so we'll wait untill we
			   get it */
			sid->current_block = alloc_dma_block(sid, GFP_ATOMIC);
#ifdef _DEBUG
			if (!sid->current_block)
				  dev_dbg(sid->dev,
					  "alloc current_block failed.\n");
#endif
		}
#endif
		_write_reg(sid, REG_STATUS_CLEAR, REG_STATUS_CLEAR_DMA_DONE);
		_write_reg(sid, REG_STATUS_CLEAR, 0);
		if (sid->current_block) {
			set_dma_address(sid, sid->current_block->handle);
			dma_start(sid);
/* 			dev_dbg(sid->dev, "dma started\n"); */
			dbg_regdump(sid);

#ifndef NO_ALLOC_IN_INTERRUPT
			sid->next_block = alloc_dma_block(sid, GFP_ATOMIC);
#ifdef _DEBUG
			if (!sid->next_block)
				dev_dbg(sid->dev,
					"alloc next_block failed.\n");
#endif
#else
			sid->next_block = NULL;
#endif
			complete(&sid->complete);
		}
		else {
			dev_err(sid->dev, "oops, no dma block\n");
		}
	}

	if (unlikely(status & REG_STATUS_OVERRUN)) {
		dev_warn(sid->dev, "hardware fifo overrun detected\n");
		sid->stat.overrun++;
		_write_reg(sid, REG_STATUS_CLEAR, REG_STATUS_CLEAR_OVERRUN);
	}

	return IRQ_HANDLED;
}

static int sid_open(struct inode *inode, struct file *filp)
{
	unsigned int minor = iminor(inode);
	struct sid *sid;
	int ret;

	/* currently we only support minor 0 */
	if (minor != 0)
		return -EINVAL;

	sid = filp->private_data = sids[minor];
	dev_dbg(sid->dev, "%s\n", __FUNCTION__);
	dev_dbg(sid->dev, "ADC Channels = 0x%x\n", sid->channels);
	dev_dbg(sid->dev, "ADC Resolution = 0x%x\n", sid->resolution);

	spin_lock_init(&sid->dma_blocks_lock);
	INIT_LIST_HEAD(&sid->dma_blocks);
	INIT_LIST_HEAD(&sid->read_blocks);
	init_completion(&sid->complete);

	ret = request_irq(sid->irq, sid_intr, SA_SHIRQ, DRIVER_NAME, sid);
	if (ret < 0) {
		dev_err(sid->dev, "Failed to obtain IRQ\n");
		return ret;
	}

	ret = do_reset(sid, default_sample_rate);

	if (ret < 0)
		free_irq(sid->irq, sid);

	return ret;
}

static int sid_release(struct inode *inode, struct file *filp)
{
	unsigned int minor = iminor(inode);
	struct sid *sid = sids[minor];

	dev_dbg(sid->dev, "%s\n", __FUNCTION__);

	hw_reset(sid);
	cleanup_dma_mem(sid);

	free_irq(sid->irq, sid);

	return 0;
}

static struct file_operations sid_fops = {
	.owner	 = THIS_MODULE,
	.read	 = sid_read,
	.open	 = sid_open,
	.ioctl	 = sid_ioctl,
	.release = sid_release,
};

struct id_struct {
	u16 id;
	u16 version;
};

struct id_struct known_boards[] __initdata = {
	{ .id = 0x5009, .version = 0x200a },
	{ },
};

static u32 __init get_board_id(void __iomem *addr)
{
	return in_be32(addr + REG_ID);
}

static int __init is_known_board(u16 id, u16 version)
{
	int ret = 0;
	int i;

	for (i=0; known_boards[i].id; i++) {
		if (known_boards[i].id == id) {
			if (known_boards[i].version == version) {
				ret = 1;
				break;
			}
		}
	}

	return ret;
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,5,0)
#define to_sid(x) ((struct sid *)dev->driver_data)

static ssize_t sid_show_reg(struct_device *dev,
			    struct device_attribute *attr, char *buf)
{
	ssize_t ret;
	struct sid *sid = to_sid(dev);

	ret = snprintf(buf, PAGE_SIZE, "control:\t0x%08x\n",
		       _read_reg(sid, REG_CONTROL));
	ret += snprintf(buf+ret, PAGE_SIZE-ret, "sample rate:\t%d\n",
			_read_reg(sid, REG_SAMPLE_RATE));
	ret += snprintf(buf+ret, PAGE_SIZE-ret, "dma address:\t0x%08x\n",
			_read_reg(sid, REG_DMA_ADDRESS));
	ret += snprintf(buf+ret, PAGE_SIZE-ret, "dma length:\t0x%08x\n",
			_read_reg(sid, REG_DMA_LENGTH));
	ret += snprintf(buf+ret, PAGE_SIZE-ret, "status:\t\t0x%08x\n",
			_read_reg(sid, REG_STATUS));
	ret += snprintf(buf+ret, PAGE_SIZE-ret, "intr e:\t\t0x%08x\n",
			_read_reg(sid, REG_INTR_E));
	return ret;
}

static ssize_t sid_show_rx(struct_device *dev,
			   struct device_attribute *attr, char *buf)
{
	struct sid *sid = to_sid(dev);
	return snprintf(buf, PAGE_SIZE, "%ld\n", sid->stat.rx);
}

static ssize_t sid_show_overrun(struct_device *dev,
				struct device_attribute *attr, char *buf)
{
	struct sid *sid = to_sid(dev);
	return snprintf(buf, PAGE_SIZE, "%ld\n", sid->stat.overrun);
}

static ssize_t
sid_show_default_sample_rate(struct_device *dev,
			     struct device_attribute *attr, char *buf)
{
	return snprintf(buf, PAGE_SIZE, "%u\n", default_sample_rate);
}

static ssize_t
sid_store_default_sample_rate(struct_device *dev,
			      struct device_attribute *attr,
			      const char *buf, size_t count)
{
	unsigned int rate;

	sscanf(buf, "%u", &rate);
	if ((rate < 0) || (SAMPLE_RATE_MAX < rate))
		return -EINVAL;

	default_sample_rate = rate;
	return strnlen(buf, PAGE_SIZE);
}


static DEVICE_ATTR(reg, S_IRUGO, sid_show_reg, NULL);
static DEVICE_ATTR(rx, S_IRUGO, sid_show_rx, NULL);
static DEVICE_ATTR(overrun, S_IRUGO, sid_show_overrun, NULL);
static DEVICE_ATTR(default_sample_rate, S_IRUGO | S_IWUSR,
		   sid_show_default_sample_rate,
		   sid_store_default_sample_rate);
#endif

#define resource_start(r)	((r)->start)
#define resource_len(r)		(((r)->end - (r)->start) + 1)

/**
 *	sid_probe - find SUZAKU AD boards and setup basic structures
 *
 *	this function is in three parts
 *	  - preamble to get enough data to check id and version
 *	  - check id and version
 *	  - struct sid initialization
 */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,5,0)
static int __init sid_probe(struct_device *dev)
#else
static int __init sid_init(void)
#endif
{
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,5,0)
	struct platform_device *pdev = to_platform_device(dev);
	struct resource *res_mem, *res_irq;
#else
	static struct pci_dev _dev = {
		.name = DRIVER_NAME,
	};
	struct_device *dev = &_dev;
#endif
	struct sid *sid;
	u32 tmp;
	u16 board_id, board_version;
	int ret = 0;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
	printk("%s (%s) : %s\n", DRIVER_NAME, DRIVER_VERSION, DRIVER_DESC);
#endif

	sid = kzalloc(sizeof(struct sid), GFP_KERNEL);
	if (!sid) {
		dev_err(dev, "Failed to obtain memory for device\n");
		ret = -ENOMEM;
		goto out;
	}

	sid->dev = dev;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,5,0)
	res_mem = platform_get_resource(pdev, IORESOURCE_MEM, 0);
	if (!res_mem) {
		dev_err(dev, "Failed to get memory region resouce\n");
		ret = -ENODEV;
		goto primary_failure;
	}

	sid->region_start  = resource_start(res_mem);
	sid->region_len    = resource_len(res_mem);
#else
	sid->region_start  = SID_BASE_ADDR;
	sid->region_len    = SID_REG_SIZE;
#endif

	if (!request_mem_region(sid->region_start, sid->region_len, DRIVER_NAME)) {
		dev_err(dev, "Failed to obtain device I/O memory region\n");
		ret = -ENOMEM;
		goto primary_failure;
	}

	sid->addr =
	  (void __iomem *)ioremap_nocache(sid->region_start, sid->region_len);
	if (!sid->addr) {
		dev_err(dev, "Failed to remap device I/O memory region\n");
		ret = -EBUSY;
		goto free_region;
	}

	tmp = get_board_id(sid->addr);
	board_id = (tmp>>16) & 0xffff;
	board_version = tmp & 0xffff;

	pr_info("  OPB-SID00 (id: %#hx; ver: %#hx)\n",
		board_id, board_version);

	if (!is_known_board(board_id, board_version)) {
		pr_info("  Unsupported ID and/or Version\n");
		goto unmap_io;
	}
	/* OK we have found a board we know, set the hardware reset
	 * bit so that we know exactly what the board state is */
	hw_reset(sid);

	/* do other things we need to do */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,5,0)
	res_irq = platform_get_resource(pdev, IORESOURCE_IRQ, 0);
	if (res_irq)
		sid->irq = resource_start(res_irq);
	else {
		dev_err(dev, "Failed to get IRQ resouce\n");
		ret = -ENODEV;
		goto unmap_io;
	}
#else
	sid->irq = SID_IRQ;
#endif
	sid->channels = _read_reg(sid, REG_ADC_CHANNELS);
	sid->resolution = _read_reg(sid, REG_ADC_RESOLUTION);
	sid->sample_clock = _read_reg(sid, REG_SAMPLE_CLOCK);
	ret = set_sample_rate(sid, default_sample_rate);
	if (ret) {
		dev_err(dev, "Failed to set sample rate\n");
		goto unmap_io;
	}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,5,0)
	sid->devno = MKDEV(SID_MAJOR, 0);
	ret = register_chrdev_region(sid->devno, 1, DRIVER_NAME);
	if (ret < 0) {
		dev_err(dev, "Failed to obtain major number\n");
		goto unmap_io;
	}

	cdev_init(&sid->cdev, &sid_fops);
	sid->cdev.owner = THIS_MODULE;

	ret = cdev_add(&sid->cdev, sid->devno, 1);
	if (ret < 0) {
		dev_err(dev, "Failed to register as a character device\n");
		goto unregister;
	}

	device_create_file(dev, &dev_attr_reg);

	device_create_file(dev, &dev_attr_rx);
	device_create_file(dev, &dev_attr_overrun);
	device_create_file(dev, &dev_attr_default_sample_rate);
#else
	sid->devno = register_chrdev (SID_MAJOR, DRIVER_NAME, &sid_fops);
	if (sid->devno < 0) {
		dev_err(dev, "Failed to register as a character device\n");
		goto unmap_io;
	}
#endif

	dev->driver_data = sid;
	sids[0] = sid;

	return ret;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,5,0)
unregister:
	unregister_chrdev_region(sid->devno, 1);
#endif
unmap_io:
	iounmap(sid->addr);
free_region:
	release_mem_region(sid->region_start, sid->region_len);
primary_failure:
	kfree(sid);
out:
	return ret;
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,5,0)
static int sid_remove(struct_device *dev)
#else
static void __exit sid_cleanup(void)
#endif
{
	struct sid *sid = sids[0];

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,5,0)
	device_remove_file(dev, &dev_attr_default_sample_rate);
	device_remove_file(dev, &dev_attr_overrun);
	device_remove_file(dev, &dev_attr_rx);

	device_remove_file(dev, &dev_attr_reg);

	cdev_del(&sid->cdev);
	unregister_chrdev_region(sid->devno, 1);
#else
	unregister_chrdev (sid->devno, DRIVER_NAME);
#endif

	iounmap(sid->addr);
	release_mem_region(sid->region_start, sid->region_len);

	kfree(sid);

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,5,0)
	return 0;
#endif
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,5,0)
static struct device_driver sid_device_driver = {
	.name	= DRIVER_NAME,
	.bus	= &platform_bus_type,
	.probe	= sid_probe,
	.remove = sid_remove,
};

static int __init sid_init(void)
{
	printk("%s (%s) : %s\n", DRIVER_NAME, DRIVER_VERSION, DRIVER_DESC);

	return driver_register(&sid_device_driver);
}

static void __exit sid_cleanup(void)
{
	driver_unregister(&sid_device_driver);
}
#endif

module_init(sid_init);
module_exit(sid_cleanup);

MODULE_AUTHOR("Chris McHarg");
MODULE_LICENSE("GPL v2");
MODULE_VERSION("1.0");
MODULE_DESCRIPTION("SUZAKU I/O Boards -A/D- Driver");
